@@ -1,7 +1,5 @@
 /* Command line parsing.
-   Copyright (C) 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004,
-   2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Free
-   Software Foundation, Inc.
+   Copyright (C) 1996-2015, 2018 Free Software Foundation, Inc.
 
 This file is part of GNU Wget.
 
@@ -37,7 +35,7 @@ as that of the covered work.  */
 #include <string.h>
 #include <signal.h>
 #include <spawn.h>
-#ifdef ENABLE_NLS
+#if defined(ENABLE_NLS) || defined(WINDOWS)
 # include <locale.h>
 #endif
 #include <assert.h>
@@ -66,6 +64,12 @@ as that of the covered work.  */
 #include <getpass.h>
 #include <quote.h>
 
+#ifdef TESTING
+/* Rename the main function so we can have a main() in fuzzing code
+   and call the original main. */
+# define main main_wget
+#endif
+
 #ifdef HAVE_METALINK
 # include <metalink/metalink_parser.h>
 # include "metalink.h"
@@ -74,6 +78,9 @@ as that of the covered work.  */
 #ifdef WINDOWS
 # include <io.h>
 # include <fcntl.h>
+#ifndef ENABLE_NLS
+# include <mbctype.h>
+#endif
 #endif
 
 #ifdef __VMS
@@ -108,7 +115,6 @@ const char *exec_name;
 /* Number of successfully downloaded URLs */
 int numurls = 0;
 
-#ifndef TESTING
 /* Initialize I18N/L10N.  That amounts to invoking setlocale, and
    setting up gettext's message catalog using bindtextdomain and
    textdomain.  Does nothing if NLS is disabled or missing.  */
@@ -148,6 +154,14 @@ i18n_initialize (void)
   /* Set the text message domain.  */
   bindtextdomain ("wget", LOCALEDIR);
   textdomain ("wget");
+#elif defined WINDOWS
+  char MBCP[16] = "";
+  int CP;
+
+  CP = _getmbcp(); /* Consider it's different from default. */
+  if (CP > 0)
+    snprintf(MBCP, sizeof(MBCP), ".%d", CP);
+  setlocale(LC_ALL, MBCP);
 #endif /* ENABLE_NLS */
 }
 
@@ -158,16 +172,12 @@ hsts_store_t hsts_store;
 static char*
 get_hsts_database (void)
 {
-  char *home;
-
   if (opt.hsts_file)
     return xstrdup (opt.hsts_file);
 
-  home = home_dir ();
-  if (home)
+  if (opt.homedir)
     {
-      char *dir = aprintf ("%s/.wget-hsts", home);
-      xfree(home);
+      char *dir = aprintf ("%s/.wget-hsts", opt.homedir);
       return dir;
     }
 
@@ -213,6 +223,7 @@ save_hsts (void)
         }
 
       hsts_store_close (hsts_store);
+      xfree (hsts_store);
 
       xfree (filename);
     }
@@ -381,6 +392,7 @@ static struct cmdline_option option_data[] =
     { "preferred-location", 0, OPT_VALUE, "preferredlocation", -1 },
 #endif
     { "preserve-permissions", 0, OPT_BOOLEAN, "preservepermissions", -1 },
+    { IF_SSL ("ciphers"), 0, OPT_VALUE, "ciphers", -1 },
     { IF_SSL ("private-key"), 0, OPT_VALUE, "privatekey", -1 },
     { IF_SSL ("private-key-type"), 0, OPT_VALUE, "privatekeytype", -1 },
     { "progress", 0, OPT_VALUE, "progress", -1 },
@@ -488,8 +500,14 @@ static unsigned char optmap[96];
 static void
 init_switches (void)
 {
+  static bool initialized;
   char *p = short_options;
   size_t i, o = 0;
+
+  if (initialized)
+    return;
+  initialized = 1;
+
   for (i = 0; i < countof (option_data); i++)
     {
       struct cmdline_option *cmdopt = &option_data[i];
@@ -551,10 +569,14 @@ init_switches (void)
 
 /* Print the usage message.  */
 static int
-print_usage (int error)
+print_usage (_GL_UNUSED int error)
 {
+#ifndef TESTING
   return fprintf (error ? stderr : stdout,
                   _("Usage: %s [OPTION]... [URL]...\n"), exec_name);
+#else
+  return 0;
+#endif
 }
 
 /* Print the help message, describing all the available options.  If
@@ -562,6 +584,7 @@ print_usage (int error)
 _Noreturn static void
 print_help (void)
 {
+#ifndef TESTING
   /* We split the help text this way to ease translation of individual
      entries.  */
   static const char *help[] = {
@@ -627,6 +650,8 @@ Download:\n"),
   -t,  --tries=NUMBER              set number of retries to NUMBER (0 unlimits)\n"),
     N_("\
        --retry-connrefused         retry even if connection is refused\n"),
+    N_("\
+       --retry-on-http-error=ERRORS    comma-separated list of HTTP errors to retry\n"),
     N_("\
   -O,  --output-document=FILE      write documents to FILE\n"),
     N_("\
@@ -768,7 +793,7 @@ HTTP options:\n"),
        --header=STRING             insert STRING among the headers\n"),
 #ifdef HAVE_LIBZ
     N_("\
-       --compression=TYPE          choose compression, one of auto, gzip and none\n"),
+       --compression=TYPE          choose compression, one of auto, gzip and none. (default: none)\n"),
 #endif
     N_("\
        --max-redirect              maximum redirections allowed per page\n"),
@@ -851,6 +876,10 @@ HTTPS (SSL/TLS) options:\n"),
        --egd-file=FILE             file naming the EGD socket with random data\n"),
 #endif
     "\n",
+    N_("\
+       --ciphers=STR           Set the priority string (GnuTLS) or cipher list string (OpenSSL) directly.\n\
+                                   Use with care. This option overrides --secure-protocol.\n\
+                                   The format and syntax of this string depend on the specific SSL/TLS engine.\n"),
 #endif /* HAVE_SSL */
 
 #ifdef HAVE_HSTS
@@ -996,7 +1025,8 @@ Recursive accept/reject:\n"),
     N_("\
   -np, --no-parent                 don't ascend to the parent directory\n"),
     "\n",
-    N_("Mail bug reports and suggestions to <bug-wget@gnu.org>\n")
+    N_("Email bug reports, questions, discussions to <bug-wget@gnu.org>\n"),
+    N_("and/or open issues at https://savannah.gnu.org/bugs/?func=additem&group=wget.\n")
   };
 
   size_t i;
@@ -1010,7 +1040,7 @@ Recursive accept/reject:\n"),
   for (i = 0; i < countof (help); i++)
     if (fputs (_(help[i]), stdout) < 0)
       exit (WGET_EXIT_IO_FAIL);
-
+#endif /* TESTING */
   exit (WGET_EXIT_SUCCESS);
 }
 
@@ -1047,7 +1077,12 @@ prompt_for_password (void)
     fprintf (stderr, _("Password for user %s: "), quote (opt.user));
   else
     fprintf (stderr, _("Password: "));
+#ifndef TESTING
+  /* gnulib's getpass() uses static variables internally, bad for fuzing */
   return getpass("");
+#else
+  return xstrdup("");
+#endif
 }
 
 
@@ -1308,6 +1343,8 @@ There is NO WARRANTY, to the extent permitted by law.\n"), stdout) < 0)
 
 const char *program_name; /* Needed by lib/error.c. */
 const char *program_argstring; /* Needed by wget_warc.c. */
+struct ptimer *timer;
+int cleaned_up;
 
 int
 main (int argc, char **argv)
@@ -1321,7 +1358,9 @@ main (int argc, char **argv)
   bool noconfig = false;
   bool append_to_log = false;
 
-  struct ptimer *timer = ptimer_new ();
+  cleaned_up = 0; /* do cleanup later */
+
+  timer = ptimer_new ();
   double start_time = ptimer_measure (timer);
 
   total_downloaded_bytes = 0;
@@ -1366,6 +1405,7 @@ main (int argc, char **argv)
 
   /* Load the hard-coded defaults.  */
   defaults ();
+  opt.homedir = home_dir();
 
   init_switches ();
 
@@ -1407,7 +1447,8 @@ main (int argc, char **argv)
 
   /* If the user did not specify a config, read the system wgetrc and ~/.wgetrc. */
   if (noconfig == false && use_userconfig == false)
-    initialize ();
+    if ((ret = initialize ()))
+      return ret;
 
   opterr = 0;
   optind = 0;
@@ -1529,6 +1570,9 @@ main (int argc, char **argv)
 
   nurl = argc - optind;
 
+  /* Initialize logging ASAP.  */
+  log_init (opt.lfilename, append_to_log);
+
   /* If we do not have Debug support compiled in AND Wget is invoked with the
    * --debug switch, instead of failing, we silently turn it into a no-op. For
    *  this no-op, we explicitly set opt.debug to false and hence none of the
@@ -1634,8 +1678,8 @@ for details.\n\n"));
            {
               /* Check if output file exists; if it does, exit. */
               logprintf (LOG_VERBOSE,
-                         _("File `%s' already there; not retrieving.\n"),
-                         opt.output_document);
+                         _("File %s already there; not retrieving.\n"),
+                         quote (opt.output_document));
               exit (WGET_EXIT_GENERIC_ERROR);
            }
     }
@@ -1725,12 +1769,14 @@ for details.\n\n"));
       )
     {
       /* No URL specified.  */
+#ifndef TESTING
       fprintf (stderr, _("%s: missing URL\n"), exec_name);
       print_usage (1);
       fprintf (stderr, "\n");
       /* #### Something nicer should be printed here -- similar to the
          pre-1.5 `--help' page.  */
       fprintf (stderr, _("Try `%s --help' for more options.\n"), exec_name);
+#endif
       exit (WGET_EXIT_GENERIC_ERROR);
     }
 
@@ -1825,13 +1871,13 @@ for details.\n\n"));
   if (opt.enable_iri)
     {
       if (opt.locale && !check_encoding_name (opt.locale))
-        opt.locale = NULL;
+        xfree (opt.locale);
 
       if (!opt.locale)
         opt.locale = find_locale ();
 
       if (opt.encoding_remote && !check_encoding_name (opt.encoding_remote))
-        opt.encoding_remote = NULL;
+        xfree (opt.encoding_remote);
     }
 #else
   memset (&dummy_iri, 0, sizeof (dummy_iri));
@@ -1865,9 +1911,14 @@ for details.\n\n"));
   if (opt.wdebug)
      dbug_init();
   sock_init();
-#else
+#elif ! defined TESTING
   if (opt.background)
-    fork_to_background ();
+    {
+      bool logfile_changed = fork_to_background ();
+
+      if (logfile_changed)
+        log_init (opt.lfilename, append_to_log);
+    }
 #endif
 
   /* Initialize progress.  Have to do this after the options are
@@ -1892,9 +1943,6 @@ for details.\n\n"));
     }
   url[i] = NULL;
 
-  /* Initialize logging.  */
-  log_init (opt.lfilename, append_to_log);
-
   /* Open WARC file. */
   if (opt.warc_filename != 0)
     warc_init ();
@@ -2203,7 +2251,7 @@ only if outputting to a regular file.\n"));
       char *wall_time = xstrdup (secs_to_human_time (end_time - start_time));
       char *download_time = xstrdup (secs_to_human_time (total_download_time));
 
-      ptimer_destroy (timer);
+      ptimer_destroy (timer); timer = NULL;
 
       logprintf (LOG_NOTQUIET,
                  _("FINISHED --%s--\nTotal wall clock time: %s\n"
@@ -2240,8 +2288,6 @@ only if outputting to a regular file.\n"));
   exit (get_exit_status ());
 }
 
-#endif /* TESTING */
-
 /*
  * vim: et ts=2 sw=2
  */
